#include "cuda.h"
#include "cuda_runtime_api.h"
#include "gdrapi.h"
#include "common.h"

/* Kernels to fill array with size int/floats of value */
__global__ void fill_array(int* array, int value, size_t size);
__global__ void fill_float_array(float* array, float value, size_t size);

/* Functions to compute average and deviation of an array.
 * Only get_average and get_reduction are to be used by main.
 * Works with square grids/blocks
 */
extern __shared__ float _temp[];
__global__ void get_reduction(float* array, float* output, size_t size, size_t stride);
__global__ void get_average_reduction(float* array, float* average, size_t size, size_t stride, size_t divider);
__global__ void get_dispersion_reduction(float* array, float* average, float* output, size_t size, size_t stride, size_t divider);
void get_average(float* array, float* average, size_t size,dim3 blocks_per_grid, dim3 threads_per_block);
void get_dispersion(float* array, float* average, float* dispersion, size_t size, dim3 blocks_per_grid, dim3 threads_per_block);

/* Some kernels operating on matrix with various methods */
__global__ void fill_matrix(int* M,int value);
__global__ void add_matrix(int* A,int* B,int* C);
__global__ void fill_matrix2(int* M,int value,size_t flat_size);
__global__ void add_matrix2(int* A,int* B,int* C,size_t flat_size);
__global__ void add_matrix_mod(int* B,int* C);

/* Fills A as an identity matrix */
__global__ void kern_identity_matrix(matrix A);

/* An implementation of parallel matrix multiplication */
__global__ void kern_mult_matrix_naive(matrix A, matrix B, matrix R);

/* A (partially) optmisation of parallel matrix multiplication.
 * Only works with square matrix of size multiple of BLOCK_SIZE
 */
__global__ void kern_mult_matrix_shared(matrix A, matrix B, matrix R);
