#include <stdio.h>
#include <stdlib.h>
#include "common.h"
#include <unistd.h>
#include <sys/time.h>
#include <math.h>
#include "kernels.h"

#define ARRAY_SIZE 224*128*16*2
#define VALUE 2.

int main(int argc, char* argv[])
{
    assert_cuda( cudaSetDevice(1) );
    cudaDeviceProp prop;
    assert_cuda( cudaGetDeviceProperties(&prop,0) );
    deviceInformation(0);
    printf("%s\n\n",prop.name);

    cudaError_t err;
    float input[ARRAY_SIZE];
    dim3 blocks_per_grid(224,1,1);
    dim3 threads_per_block(128,1,1);
    float* dev_array;

    cudaEvent_t start,stop;
    float ms;
    assert_cuda( cudaEventCreate(&start) );
    assert_cuda( cudaEventCreate(&stop) );

    assert_cuda( cudaMalloc((void**)&dev_array,ARRAY_SIZE*sizeof(float)) );

    assert_cuda( cudaEventRecord(start) );
    fill_float_array<<< blocks_per_grid,threads_per_block >>>(dev_array,VALUE,ARRAY_SIZE);
    err = cudaGetLastError();
    printf("%s: %s\n",cudaGetErrorName(err),cudaGetErrorString(err));

    assert_cuda( cudaMemcpy(input,dev_array,ARRAY_SIZE*sizeof(float),cudaMemcpyDeviceToHost) );

    float average=0.;
    float* dev_average;
    assert_cuda( cudaMalloc((void**)&dev_average,sizeof(float)) );
    /* blocks_per_grid.x=16*14; */
    /* threads_per_block.x=128; */

    assert_cuda( cudaMemcpy(dev_array,input,ARRAY_SIZE*sizeof(float),cudaMemcpyHostToDevice) );

    for(int k=0;k<ITERATION;k++)
    {
	assert_cuda( cudaEventRecord(start) );
	get_average(dev_array,dev_average,ARRAY_SIZE,blocks_per_grid,threads_per_block);
	err = cudaGetLastError();
	printf("%s: %s\n",cudaGetErrorName(err),cudaGetErrorString(err));
	assert_cuda( cudaDeviceSynchronize() );
	assert_cuda( cudaEventRecord(stop) );
	assert_cuda( cudaEventSynchronize(stop) );
	assert_cuda( cudaEventElapsedTime(&ms,start,stop) );
	assert_cuda( cudaMemcpy(&average,dev_average,sizeof(float),cudaMemcpyDeviceToHost) );
	printf("avg_gpu = %f\n",average);
	printf("Elapsed time: %f ms\n",ms);
    }


    float dispersion;
    float* dev_dispersion;
    assert_cuda( cudaMalloc((void**)&dev_dispersion,sizeof(float)) );

    for(int k=0;k<ITERATION;k++)
    {
	assert_cuda( cudaEventRecord(start) );
	get_dispersion(dev_array,dev_average,dev_dispersion,ARRAY_SIZE,blocks_per_grid,threads_per_block);
	err = cudaGetLastError();
	printf("%s: %s\n",cudaGetErrorName(err),cudaGetErrorString(err));
	assert_cuda( cudaDeviceSynchronize() );
	assert_cuda( cudaEventRecord(stop) );
	assert_cuda( cudaEventSynchronize(stop) );
	assert_cuda( cudaEventElapsedTime(&ms,start,stop) );
	assert_cuda( cudaMemcpy(&average,dev_average,sizeof(float),cudaMemcpyDeviceToHost) );
	assert_cuda( cudaMemcpy(&dispersion,dev_dispersion,sizeof(float),cudaMemcpyDeviceToHost) );
	printf("avg_gpu = %f\n",average);
	printf("disp_gpu = %f\n",dispersion);
	printf("Elapsed time: %f ms\n",ms);
    }


    struct timeval t1,t2;
    float time;

    for(int k=0;k<ITERATION;k++)
    {
	gettimeofday(&t1,0);
	average = cpu_average(input,ARRAY_SIZE);
//	dispersion = cpu_dispersion(input,average,ARRAY_SIZE);
	gettimeofday(&t2,0);
	time = t2.tv_usec - t1.tv_usec;
	printf("avg_cpu = %f\n",average);
	printf("disp_cpu = %f\n",dispersion);
	printf("Elapsed time: %f us\n",time);
    }

    assert_cuda( cudaEventDestroy(start) );
    assert_cuda( cudaEventDestroy(stop) );
    assert_cuda( cudaFree(dev_average) );
    assert_cuda( cudaFree(dev_dispersion) );
    assert_cuda( cudaFree(dev_array) );
    exit( EXIT_SUCCESS );
}
